-- general functions for invaders game

--------------------------------------------------------------------------------
-- Removes item val from table t
function remove_item(t, val)
  for i,v in ipairs(t) do
    if v == val then
      table.remove(t, i)
      break
    end
  end
end

--------------------------------------------------------------------------------
function is_in_table(t, item)
  local result = false
  for i=1,#t do
    if item == t[i] then
      result = true
      break
    end
  end
  
  return result
end


--------------------------------------------------------------------------------
-- returns a random time interval (in milliseconds) based on the
-- poisson process
function poisson_interval(rate)
  return (-math.log(1-math.random())/rate)*1000
end



--------------------------------------------------------------------------------
-- returns a table of characters of the string s
function str_to_chars(s)
  chars = {}
  for i=1,#s do
    chars[i] = string.sub(s, i, i)
  end
  return chars
end



--------------------------------------------------------------------------------
-- returns a table of decimal r, g, b values from a 6 digit hex string
function hex_to_rgb(hex)
  local color = {}
  local digits = str_to_chars(hex)
  
  color.r = tonumber(digits[1]..digits[2], 16)
  color.g = tonumber(digits[3]..digits[4], 16)
  color.b = tonumber(digits[5]..digits[6], 16)
  
  return color
end



--------------------------------------------------------------------------------
-- returns a string representing the percent of a number
-- 0.8376658 -> 83%
function format_percent(number)
  return string.format("%d", number*100).."%"
end


--------------------------------------------------------------------------------
-- returns a string representing a time t (t in milliseconds)
-- 2384 -> 2.38s
function format_time(t)
  return string.format("%.2f", t/1000).."s"
end



--------------------------------------------------------------------------------
-- returns linear interpolation point from point p in direction of vector v
-- t = [0, 1]
function lerp(p, v, t)
  if t <= 0 then
    return p
  end
  
  if t >= 1 then
    local x = p.x + v.x
    local y = p.y + v.y
    return vect:new(x, y)
  end
  
  local x = p.x + v.x * t
  local y = p.y + v.y * t
  
  return vect:new(x, y)
end


--------------------------------------------------------------------------------
function get_bounce_velocity(velocity, normal, friction, restitution)
  if normal.x == nil or normal.y == nil then
    return vector2:new(0,0)
  end
  local v_dot_n = vector2:dot(velocity, normal)
  local u = v_dot_n * normal
  local w = velocity - u
  local new_velocity = friction * w - restitution * u
  
  return new_velocity
end 


--------------------------------------------------------------------------------
-- returns table of supported resolutions
function get_supported_resolutions(max_width, max_height)

  -- table of common resolutions:
  local resolutions = { {800, 600},
                        {1024, 768},
                        {1280, 800},
                        {1366, 768},
                        {1440, 900},
                        {1600, 900},
                        {1680, 1050},
                        {1920, 1080} }
                 
  -- remove resolutions greater than max
  for i = #resolutions,1,-1 do
    local res = resolutions[i]
    if res[1] > max_width or res[2] > max_height then
      table.remove(resolutions, i)
    end
  end
  
  -- if max resolution is not in list, then add it in
  local is_in_table = false
  for _,v in ipairs(resolutions) do
    if v[1] == max_width and v[2] == max_height then
      is_in_table = true
      break
    end
  end
  
  if not is_in_table then
    table.insert(resolutions, 1, {max_width, max_height})
  end
  
  -- sort table
  table.sort(resolutions, function(a,b) 
                            if a[2] == b[2] then
                              return a[1]>b[1] 
                            else
                              return a[2]>b[2] 
                            end
                          end)
  
  return resolutions
end





























