
--##########################################################################--
--[[----------------------------------------------------------------------]]--
-- hero object
--[[----------------------------------------------------------------------]]--
--##########################################################################--
local hero = {}
hero.table = HERO
hero.debug = false
hero.map = nil

-- input
hero.click_timer = nil              
hero.dclick_time = 200              -- in milliseconds
hero.dclick_effect_time = 500       -- in milliseconds
hero.dclick_timer = nil             -- timer for double click effect
hero.dclick_dir = nil               -- direction when mouse was double clicked
hero.input_radius = 150             -- pixels, mouse boundary radius
hero.input_direction = nil
hero.input_power = 0
hero.thrust_time = 250              -- time for thrust to reach full power
hero.input_thrust = 0               -- [0,1], thrust value

-- direction
hero.direction = vector2:new(0, -1) -- current direction
hero.num_dirs = 30                  -- number of past direcitons to remember
hero.dir_history = nil							-- array of past num_dirs directions

-- movement
hero.point = nil
hero.pos = VECT_ZERO
hero.max_force = 12
hero.max_velocity = 10
hero.min_velocity = 3
hero.velocity_limit = hero.max_velocity
hero.friction = 1.5

-- camera
hero.camera_point = nil
hero.camera_target = VECT_ZERO
hero.camera = nil
hero.cam_radius = 0          -- current dist camera target is away from hero
hero.act_cam_radius = 0      -- actual radius (target radius for cam_radius)
hero.cam_max_radius = 180
hero.shrink_speed = 8
hero.enlarge_speed = 40
hero.min_approach = 0.4      -- approaching slowing factor for camera arriving
hero.max_approach = 1.5      -- at the camera target

-- collision
hero.bbox = nil
hero.bbox_width = 15
hero.bbox_height = 15

-- weapon
hero.gun = nil

hero.m_point = nil

local hero_mt = { __index = hero }
function hero:new(pos, map, camera)
	local point = physics.point:new(pos)
	
	-- input timers
	local click_timer = timer:new()
	local dclick_timer = timer:new(hero.dclick_effect_time)
	
	-- direction
	local dir_history = {}
	for i=1,hero.num_dirs do
		dir_history[i] = hero.direction
	end
	
	--camera
	local target = vector2:new(SCR_WIDTH/2, SCR_HEIGHT/2)
	local cam_point = physics.steer:new()
	cam_point:set_target(target)
	cam_point:set_position(target)
	cam_point:set_mass(0.25)
	cam_point:set_force(5)
	cam_point:set_max_speed(20)
  cam_point:set_radius(300)
	
  -- bounding box
  local bx = pos.x - 0.5 * hero.bbox_width
  local by = pos.y - 0.5 * hero.bbox_height
  local bbox = bbox:new(bx, by, hero.bbox_width, hero.bbox_height)
  collider:add_object(bbox)
  
  -- body
  local points = {vector2:new(bbox.x, bbox.y),
									vector2:new(bbox.x + bbox.width, bbox.y),
									vector2:new(bbox.x + bbox.width, bbox.y + bbox.height),
									vector2:new(bbox.x, bbox.y + bbox.height)}
  local body = map_body:new(points, map)
  body:set_origin(pos)
  
  local m_point = map_point:new(pos, map)
  
  hero_gun = gun:new(pos, hero.direction, map)
  
  local hero_table = setmetatable({ point = point,
																		click_timer = click_timer,
																		dclick_timer = dclick_timer,
																		dir_history = dir_history,
																		camera_point = cam_point,
																		bbox = bbox,
																		map = map,
																		camera = camera,
																		m_point = m_point,
	                                  gun = hero_gun,
	                                  body = body}, hero_mt)
	
	body:set_parent(hero_table)
	bbox.parent = hero_table
	hero_gun:set_parent(hero_table)
  
  return hero_table
end

function hero:shoot()
	self.gun:shoot()
end

------------------------------------------------------------------------------
function hero:update(dt)
	self:_update_input(dt)
	self:_update_direction(dt)
	self:_update_position(dt)
	self:_update_camera_target(dt)
	self:_update_collision(dt);
	self:_update_weapon(dt)
end

------------------------------------------------------------------------------
function hero:_update_weapon(dt)
	-- set orientation
	local dir = self.direction
	local pos = self.pos + 10 * dir
	self.gun:set_position(pos)
	self.gun:set_direction(dir)
	self.gun:update(dt)
end

------------------------------------------------------------------------------
function hero:_update_collision(dt)
	-- update position of bbox
	local bbox = self.bbox
	local pos = self:get_position()
  bbox.x =  pos.x - 0.5 * hero.bbox_width
  bbox.y = pos.y - 0.5 * hero.bbox_height
  collider:update_object(bbox.cdata.id)
  
  -- update position of map point
  self.m_point:set_position(pos)
  self.m_point:update(dt)
  local c, n, p = self.m_point:get_collision_data()
  if c then
  	self.point.vel = get_bounce_velocity(self.point.vel, n, 0.9, 0.3)
  	self:set_position(p)
  end
  
  -- update position of body
  self.body:set_position(self:get_position())
  self.body:update(dt)
  local c, n, p, offset = self.body:get_collision_data()
  if c then
  	local new_pos = self.body:get_position() + offset
  	local success, safe_pos = self.body:update_position(new_pos)
  	if not success then
  		new_pos = safe_pos
  	end
  	
  	self:set_position(safe_pos)
  	self.point.vel = get_bounce_velocity(self.point.vel, n, 0.9, 0.3)
  end
  
end

------------------------------------------------------------------------------
function hero:_update_position(dt)
	-- calculate pushing force, limit max_velocity
	self.limit_velocity = self.max_velocity
	if self.input_direction then
		local thrust = self.input_thrust
		local power = self.input_power
		local max_force = self.max_force
		local dir = self.input_direction

		local force = thrust * power * max_force * dir
		self.point:add_force(force)
	end

	-- set velocity limit
	local vmag = self.point:get_velocity():mag()
	if self.input_power == 0 then
		self.limit_velocity = self.max_velocity
	else
		local min, max = self.min_velocity, self.max_velocity
		local limit = math.max(min + self.input_power * (max - min), vmag)
		limit = math.min(limit, self.max_velocity)
		self.limit_velocity = limit
	end
	
	-- limit velocity
	local vdir = self.point:get_velocity():normalize()
	if vmag > self.limit_velocity then
		self.point:set_velocity(self.limit_velocity * vdir)
	end
	
	-- calculate friction force (only when idle)
	if self.input_power == 0 then
		local vmag = self.point:get_velocity():mag()
		local force_fr = -self.friction * vdir *  vmag
		self.point:add_force(force_fr)
	else
		local vmag = self.point:get_velocity():mag()
		local force_fr = -self.friction * vdir *  vmag / 4
		self.point:add_force(force_fr)
	end
	
	self.point:update(dt)
	self.pos = self.point:get_position()
	
end

------------------------------------------------------------------------------
function hero:_update_direction(dt)
	local next_dir = self.input_direction or self.direction
	local past_dirs = self.dir_history
	
	-- remove value at head, and insert new at tail
	table.remove(past_dirs, 1)
	past_dirs[self.num_dirs] = next_dir
	
	-- find average direction
	local avg = vector2:new(0, 0)
	for i=1,#past_dirs do
		local dir = past_dirs[i]
		avg = avg + dir
	end
	self.direction = avg:normalize()
end

------------------------------------------------------------------------------
function hero:_update_input()
	-- check whether mouse has been clicked/double clicked
	local double_clicked = false
	local clicked = controls.click:was_pressed()
	if clicked then
		local time_passed = self.click_timer:time_elapsed()
	
		self.click_timer:reset()
		self.click_timer:start()
		self.dclick_timer:reset()    -- cancels double click effect
		
		if time_passed and time_passed < self.dclick_time then
			double_clicked = true
			self.click_timer:reset()
		end
	end

	-- start double clicked effect timer and set direction
	if double_clicked then
		self.dclick_timer:start()
		local dclick_dir = mouse:get_position() - self:get_screen_position()
		self.dclick_dir = dclick_dir:normalize()
	end
	
	-- make sure the mouse position doesnt move with the screen
	local mouse_pos = mouse:get_position()
	local screen_pos = self:get_screen_position()
	if self.input_length and controls.click:is_down() then
		local mouse_diff = mouse_pos - self.last_mouse_pos
		local new_pos = screen_pos + self.input_direction * self.input_length + mouse_diff
		mouse:set_position(new_pos)
		mouse_pos = new_pos
	end
	self.last_mouse_pos = mouse_pos
	
	-- calc direction/power of input
	local direction, length
	local power = 0
	if controls.click:is_down() then		
		direction = mouse_pos - screen_pos
		length = direction:mag()
		direction = direction:normalize()	
		
		-- limit mouse position
		if length > self.input_radius then
			local new_pos = screen_pos + self.input_radius * direction
			--mouse:set_position(new_pos)
			length = self.input_radius
		end
		
		power = length / self.input_radius
	end
	
	if not direction and self.dclick_timer:isrunning() then
		direction = self.dclick_dir
	end
	
	-- calculate thrust
	local time = controls.click.time_pressed * 1000
	local thrust = 1
	if time < self.thrust_time then
		thrust = time / self.thrust_time
	end
	
	self.input_direction = direction
	self.input_power = power
	self.input_length = length
	self.input_thrust = thrust
	
	--joystick test
	if love.joystick.isOpen(1) then
		self.input_direction = jdir
		self.input_power = jpower
		self.input_length = length
		self.input_thrust = 1
	end
	
end

------------------------------------------------------------------------------
function hero:_update_camera_target(dt)
	-- calc camera radius
	local vmag = self.point:get_velocity():mag()
	local vratio = vmag / self.max_velocity
	local target_radius = vratio * self.cam_max_radius
	
	-- set radius based on target radius
	local radius = self.cam_radius
	if radius < target_radius then
		radius = math.min(radius + self.enlarge_speed * dt, target_radius)
	else
		radius = math.max(radius - self.shrink_speed * dt, target_radius)
	end
	self.cam_radius = radius

	-- update approaching behavior of camera depending on player speed
	local cam_target = self.camera.target
	local vratio = self.point.vel:mag() / self.max_velocity
	local minsf, maxsf = self.min_approach, self.max_approach
	cam_target:set_approach_factor(minsf + vratio * (maxsf - minsf))
	
	-- place camera target on the camera radius in direction of player
	self.camera_target = self.point.pos + radius * self.direction
end

------------------------------------------------------------------------------
function hero:draw()
	lg.setColor(C_RED)
	lg.setLine(2)
	lg.circle('line', self.pos.x, self.pos.y, 5)
	
	local p1 = self.pos
	local p2 = p1 + 15 * self.direction
	lg.line(p1.x, p1.y, p2.x, p2.y)
	
	self.gun:draw()
	
	if self.debug then
		lg.setColor(0, 255, 0, 255)
		lg.setLine(1)
		lg.circle('line', self.pos.x, self.pos.y, self.cam_max_radius)
		lg.setColor(0, 255, 255, 255)
		lg.circle('line', self.pos.x, self.pos.y, self.cam_radius)
		
		local w = 10
		local pos = self.camera_target
		lg.setColor(0, 255, 0, 255)
		lg.rectangle('line', pos.x - w/2, pos.y - w/2, w, w)
		
		-- bbox
		lg.setColor(0, 200, 0, 255)
		self.bbox:draw()
		
		--weapon
		self.gun:draw_debug()
		
		-- body
		self.body:draw()
	end
end

------------------------------------------------------------------------------
function hero:get_camera_target()
	return self.camera_target
end

function hero:get_screen_position()
	return (self.pos - self.camera:get_pos()) * self.camera.scale
end

function hero:get_position()
	return self.pos
end

function hero:set_position(pos)
	self.pos = pos
	self.point:set_position(pos)
	self.m_point:update_position(pos)
	
	local bbox = self.bbox
	local pos = self:get_position()
  bbox.x =  pos.x - 0.5 * hero.bbox_width
  bbox.y = pos.y - 0.5 * hero.bbox_height
  collider:update_object(bbox.cdata.id)
end

function hero:set_camera(cam)
	self.camera = camera
end

function hero:get_velocity()
	return self.point:get_velocity()
end

return hero



