
--##########################################################################--
--[[----------------------------------------------------------------------]]--
-- map_body object
--[[----------------------------------------------------------------------]]--
--##########################################################################--
local map_body = {}
map_body.table = MAP_BODY
map_body.debug = true
map_body.parent = nil
map_body.iscurrent = true
map_body.mpoints = nil
map_body.next_points = nil
map_body.old_points = nil
map_body.old_origin = nil

-- collision
map_body.collided = false
map_body.cnormal = nil
map_body.cpoint = nil
map_body.c_mpoint = nil

local map_body_mt = { __index = map_body }
function map_body:new(points, map)
	
	-- create map points from table of points (vector2's)
	local avg_point = vector2:new(0,0)
	local mpoints = {}
	for i=1,#points do
		local p = points[i]
		local mpoint = map_point:new(p, map)
		mpoints[#mpoints+1] = mpoint
		
		avg_point = avg_point + p
	end
	
	local origin = avg_point / #points

  return setmetatable({ mpoints = mpoints,
                        origin = origin }, map_body_mt)
end

function map_body:set_parent(parent)
	self.parent = parent
end

function map_body:set_origin(pos)
	self.origin = pos
end

function map_body:translate(offset)
	if self.iscurrent then
		self.old_origin = self.origin
	end

	local next_points = self.next_points
	local mpoints = self.mpoints
	
	if next_points == nil then
		next_points = {}
		for i=1,#self.mpoints do
			next_points[i] = mpoints[i]:get_position() + offset
		end
	else
		for i=1,#next_points do
			next_points[i] = next_points[i] + offset
		end
	end
	self.next_points = next_points
	self.origin = self.origin + offset
	self.iscurrent = false
end

function map_body:set_position(new_pos)
	local offset = new_pos - self.origin
	self:translate(offset)
end
function map_body:get_position()
	return self.origin
end

function map_body:update_position(new_pos)
	local old_origin = self.old_origin or self.origin

	local offset = new_pos - self.origin
	local mpoints = self.mpoints
	local is_colliding = false
	for i=1,#mpoints do
		local m = mpoints[i]
		local new = m:get_position() + offset
		m:update_position(new)
		
		local tile = m:get_tile()
		if not tile.type.walkable then
			is_colliding = true
		end
	end
	
	self.origin = new_pos
	
	-- move to old position if new position results in collision
	-- this can happen when more then one map point is colliding at the same time
	if is_colliding then
		local offset = old_origin - self.origin
		for i=1,#mpoints do
			local m = mpoints[i]
			local new = m:get_position() + offset
			mpoints[i]:update_position(new)
		end
		
		self.origin = old_origin
	end
	
	return not is_colliding, self.origin
end

function map_body:get_collision_data()
	if self.collided then
		return true, self.cnormal, self.cpoint, self.coffset
	else
		return false
	end
end

-- true if all tiles walkable, false otherwise
function map_body:check_tiles()
	local result = true
	local mpoints = self.mpoints
	for i=1,#mpoints do
		local tile = mpoints[i]:get_tile()
		if not tile.type.walkable then
			result = false
			break
		end
	end
	
	return result
end

------------------------------------------------------------------------------
function map_body:update(dt)
	if self.iscurrent then return end
	
	-- move mpoints to new positions
	self.collided = false
	local mpoints = self.mpoints
	local next_points = self.next_points
	for i=1,#mpoints do
		mpoint = mpoints[i]
		mpoint:set_position(next_points[i])
		mpoint:update(dt)
		c, n, p, poffset = mpoint:get_collision_data()
		
		if c then
			self.collided = true
			self.cnormal = n
			self.cpoint = p
			self.c_mpoint = mpoint
			self.coffset = poffset
		end
	end
	
	self.next_points = nil
	self.iscurrent = true
end

------------------------------------------------------------------------------
function map_body:draw()
	if self.debug then
		-- draw points
		lg.setColor(0,255,0,255)
		lg.setPoint(4, "smooth")
		for i=1,#self.mpoints do
			local p = self.mpoints[i]
			local pos = p:get_position()
			lg.point(pos.x, pos.y)
		end
		
		-- origin
		lg.setColor(255, 0, 0, 255)
		lg.point(self.origin.x, self.origin.y)
		
	end
	
end

return map_body























