
--##########################################################################--
--[[----------------------------------------------------------------------]]--
-- map_point object
--[[----------------------------------------------------------------------]]--
--##########################################################################--
local map_point = {}
map_point.table = MAP_POINT
map_point.debug = false
map_point.pos = nil
map_point.next_pos = nil
map_point.map = nil
map_point.is_current = false
map_point.tile = nil              -- tile that point is on
map_point.tile_offset = nil       -- pixel offset from top left corner of tile
map_point.neighbours = nil        -- tiles surrounding tile that point is on
map_point.open_tiles = nil

-- collision data
map_point.collided = false
map_point.cpoint = nil
map_point.cnormal = nil
map_point.coffset = nil

-- used by collider
map_point.cdata = nil

local map_point_mt = { __index = map_point }
function map_point:new(pos, map)
	local tx, ty = math.floor(pos.x/TILE_WIDTH) + 1, math.floor(pos.y/TILE_HEIGHT) + 1
	local tile = map.tiles[ty][tx]
	local tile_offset = vector2:new(pos.x % TILE_WIDTH, pos.y % TILE_HEIGHT)
	local neighbours = tile.neighbours
	local open = map_point:_get_open_tiles(tile)
	
	
  return setmetatable({ pos = pos,
                        map = map,
                        tile = tile,
                        tile_offset = tile_offset,
                        neighbours = neighbours,
                        open_tiles = open }, map_point_mt)
end

------------------------------------------------------------------------------
function map_point:update(dt)
	if self.is_current then return end
	
	-- calculate new offset from next_pos
	local next_pos = self.next_pos
	local pdiff = next_pos - self.pos
	local offset = self.tile_offset
	offset = offset + pdiff
	
	-- check if offset moves point to another tile
	-- if id does, then we need to check for collision
	self.collided = false
	local change_tile = offset.x >= TILE_WIDTH or offset.y >= TILE_HEIGHT or
	                    offset.x < 0 or offset.y < 0
	if change_tile then
		local old_tile = self.tile
		local new_tile = self:_get_tile(next_pos)

		local c, n, p, off = self:_check_collision(old_tile, new_tile, self.pos, next_pos)
		if c then
			self.collided = true
			self.cnormal = n
			self.cpoint = p
			self.coffset = off
		else
			self.collided = false
		end
		
		self.tile = new_tile
		offset = self:_get_tile_offset(next_pos)
		self.neighbours = self.tile.neighbours
		self.open_tiles = self:_get_open_tiles(self.tile)
	end
	
	self.tile_offset = offset
	self.pos = next_pos
	self.is_current = true
	
end

------------------------------------------------------------------------------
function map_point:draw()
	if self.debug then
		lg.setColor(C_RED)
		lg.setPoint(4, 'smooth')
		lg.point(self.pos.x, self.pos.y)
		
		local tile = self.tile
		if tile then
			-- occupied tile
			lg.setColor(0, 255, 0, 60)
			lg.rectangle('fill', tile.x, tile.y, TILE_WIDTH, TILE_HEIGHT)
			
			-- neighbours
			local open = self.open_tiles
			for i=1,8 do
				local tile = self.neighbours[i]
				if tile then
				  if open[i] then
				  	lg.setColor(0, 255, 0, 140)
				  else
				  	lg.setColor(255, 0, 0, 140)
				  end
				  lg.rectangle('fill', tile.x, tile.y, TILE_WIDTH, TILE_HEIGHT)
				  
				end
			end
			
		end
		
		-- draw collision
		if self.cpoint then
			local p = self.cpoint
			local n = self.cnormal
			local len = 20
			local e = p + len * n
			
			lg.setColor(0, 255, 0, 255)
			lg.setPoint(3, 'smooth')
			lg.setLine(2)
			lg.point(p.x, p.y)
			lg.line(p.x, p.y, e.x, e.y)
		end
		
	end
	
end

------------------------------------------------------------------------------
--[[
 checks if there is a collision from point p_old to p_new, and
 from tile t_old to t_new
 
 returns: false if no collision
          surface normal, point of collision otherwise
]]--
function map_point:_check_collision(t_old, t_new, p_old, p_new)
	-- find which neighbour new tile is in relation to old tile
	local n = nil
	for i=1,8 do
		if t_new == t_old.neighbours[i] then
			n = i
			break
		end
	end
	
	-- can the player get to that tile?
	local open = self.open_tiles
	if open[n] then
		return false
	end
	self.collided = true
	
	-- find surface normal to collision
	local normal
	if n % 2 == 1 then                -- up, right down left
		normal = NORMAL[(n + 4) % 8]
	else                              -- diagonal
		local nb = t_old.neighbours
		local r_dir = n + 1
		local l_dir = n - 1
		if r_dir == 9 then r_dir = 1 end
		if l_dir == 0 then r_dir = 8 end
		
		walk_diag  = nb[n].type.walkable
		walk_right = nb[r_dir].type.walkable
		walk_left  = nb[l_dir].type.walkable
		
		if (not walk_right and not walk_left) or (walk_right and walk_left) then
			-- This will be a diagonal normal
			local opp = (n + 4) % 8
			if opp == 0 then opp = 8 end
			normal = NORMAL[opp]
		elseif not walk_right then
			normal = NORMAL[(r_dir + 4) % 8]
		else
			normal = NORMAL[(l_dir + 4) % 8]
		end
	end
	
	-- find point of collision
	local max, min = math.max, math.min
	local cpoint = vector2:new(0, 0)
	if n % 2 == 1 then
		-- top/right/bottom/left cases
		
		-- eq of line for calulating intersection
		local dx = max(max(p_new.x, p_old.x) - min(p_new.x, p_old.x), EPSILON)
		local dy = p_new.y - p_old.y
		local m = dy/dx
		local b = p_new.y - m * p_new.x
		if     n == UP then
			cpoint.y = t_new.y + TILE_HEIGHT
			cpoint.x = (1/m) * (cpoint.y - b)
		elseif n == RIGHT then
			cpoint.x = t_new.x
			cpoint.y = m * cpoint.x + b
		elseif n == DOWN then
			cpoint.y = t_new.y 
			cpoint.x = (1/m) * (cpoint.y - b)
		elseif n == LEFT then
			cpoint.x = t_new.x + TILE_WIDTH
			cpoint.y = m * cpoint.x + b
		end
	else
		-- diagonal cases
		if (not walk_right and not walk_left) or (walk_right and walk_left) then
			-- pure diagonal
			if     n == UPRIGHT then
				cpoint.x = t_new.x 
				cpoint.y = t_new.y + TILE_HEIGHT 
			elseif n == DOWNRIGHT then
				cpoint.x = t_new.x 
				cpoint.y = t_new.y 
			elseif n == DOWNLEFT then
				cpoint.x = t_new.x + TILE_WIDTH 
				cpoint.y = t_new.y
			elseif n == UPLEFT then
				cpoint.x = t_new.x + TILE_WIDTH 
				cpoint.y = t_new.y + TILE_HEIGHT 
			end
		else
			-- side collision diagonal
			
			-- eq of line for calulating intersection
			local dx = max(max(p_new.x, p_old.x) - min(p_new.x, p_old.x), EPSILON)
			local dy = p_new.y - p_old.y
			local m = dy/dx
			local b = p_new.y - m * p_new.x
			if     normal == NORMAL[DOWN] then   -- hit top side
				cpoint.y = t_new.y + TILE_HEIGHT
				cpoint.x = (1/m) * (cpoint.y - b)
			elseif normal == NORMAL[LEFT] then   -- hit right side
				cpoint.x = t_new.x
				cpoint.y = m * cpoint.x + b
			elseif normal == NORMAL[UP] then     -- hit bottom side
				cpoint.y = t_new.y
				cpoint.x = (1/m) * (cpoint.y - b)
			elseif normal == NORMAL[RIGHT] then  -- hit left side
				cpoint.x = t_new.x + TILE_WIDTH
				cpoint.y = m * cpoint.x + b
			end
		end
		
	end
	
	-- move the collision point back from the wall a bit
	cpoint = cpoint + 0.1*normal
	
	-- make sure the collision point is located on a walkable tile.
	-- This occurs when the point moves to a top left or downleft corner
	-- while the collision point is calculated as a pure side collision.
	-- in this rare case, p_old is on a walkable tile and is very close to the
	-- actual point of collision.
	local tile = self:_get_tile(cpoint)
	if not tile.type.walkable then
		cpoint = p_old
	end
	
	local offset = cpoint - p_new
	
	return true, normal, cpoint, offset
end

------------------------------------------------------------------------------
-- returns tile under the point p
function map_point:_get_tile(p)
	local tx = math.floor(p.x/TILE_WIDTH) + 1
	local ty = math.floor(p.y/TILE_HEIGHT) + 1
	local w, h = self.map.width, self.map.height
	
	if tx < 1 or tx > w or ty < 1 or ty > h then
		return -- out of map boundary
	end
	
	return self.map.tiles[ty][tx]
end

------------------------------------------------------------------------------
function map_point:_get_tile_offset(p)
	return vector2:new(p.x % TILE_WIDTH, p.y % TILE_HEIGHT)
end

------------------------------------------------------------------------------
-- returns table of booleans cooresponding to neighbour tiles player can move to
function map_point:_get_open_tiles(tile)
	local up, down, left, right = true, true, true, true
	local upleft, upright, downleft, downright = true, true, true, true
	local n = tile.neighbours
	
	if not n[UP].type.walkable then
		upleft, up, upright = false, false, false
	end
	if not n[DOWN].type.walkable then
		downleft, down, downright = false, false, false
	end
	if not n[LEFT].type.walkable then
		upleft, left, downleft = false, false, false
	end
	if not n[RIGHT].type.walkable then
		upright, right, downright = false, false, false
	end
	
	if not n[UPLEFT].type.walkable    then upleft = false end
	if not n[UPRIGHT].type.walkable   then upright = false end
	if not n[DOWNLEFT].type.walkable  then downleft = false end
	if not n[DOWNRIGHT].type.walkable then downright = false end
	
	return {up, upright, right, downright, down, downleft, left, upleft}
end

------------------------------------------------------------------------------
function map_point:get_collision_data()
	if self.collided then
		return true, self.cnormal, self.cpoint, self.coffset
	else
		return false
	end
end

------------------------------------------------------------------------------
function map_point:get_x() return self.pos.x end
function map_point:get_y() return self.pos.y end
function map_point:get_position() return self.pos end
function map_point:get_tile()
	return self.tile
end

------------------------------------------------------------------------------
function map_point:set_position(pos)
	self.next_pos = pos
	self.is_current = false
end

function map_point:update_position(pos)
	self.pos = pos
	self.next_pos = pos
	self.tile = self:_get_tile(pos)
	self.tile_offset = self:_get_tile_offset(pos)
	self.open_tiles = self:_get_open_tiles(self.tile)
	self.neighbours = self.tile.neighbours
	self.is_current = true
end

return map_point



