
--##########################################################################--
--[[----------------------------------------------------------------------]]--
-- tile_map object
--[[----------------------------------------------------------------------]]--
--##########################################################################--
local tile_map = {}
tile_map.table = TILE_MAP
tile_map.tiles = nil        -- 2d array of all tiles
tile_map.width = nil        -- # of tiles wide
tile_map.height = nil       -- # tiles high
tile_map.num_tiles = nil    -- total # of tiles in the tile map

-- chunk data
tile_map.ch_cols = 16        -- chunk width in # of tiles
tile_map.ch_rows = 16       -- chunk height in # of tiles
tile_map.ch_height = nil     -- # of chunks high
tile_map.ch_width = nil      -- # of chunks wide
tile_map.chunks = nil        -- 2d array of all chunks
tile_map.num_chunks = nil    -- total # of chunks

-- load data
tile_map.is_loaded = false
tile_map.is_loading = false
tile_map.load_time = 0.0167*0.75      -- allotted loading time per frame (seconds)
tile_map.load_data = nil
tile_map.load_timer = nil
tile_map.time_loading = 0

-- color
tile_map.palette = nil
tile_map.p_height = nil      -- # of colors
tile_map.p_width = nil       -- max # of shades in all colors
tile_map.p_imgdata = nil     -- holds imagedata of palette
tile_map.p_img = nil         -- generated image of the palette
tile_map.num_shades = nil    -- total # of shades in the palette
tile_map.p_quads = nil      -- contains 2d array of quads for p_img
                             -- each val corresponds to val in palette
-- tile type data
tile_map.layer_data = nil
tile_map.layer_imgs = nil
tile_map.num_layers = nil

-- camera
tile_map.camera = nil  
tile_map.map_isvisible = false
tile_map.min_x = 1
tile_map.min_y = 1
tile_map.max_x = 1
tile_map.max_y = 1

local tile_map_mt = { __index = tile_map }
function tile_map:new(width, height)
  local load_data = {}
  load_data.stage = 0
  load_data.percent = 0
  load_data.context = {}
  
  local tiles = {}
  for i=1,height do
    tiles[i] = {}
  end
  
  local ch_height = math.ceil(height/self.ch_rows)
  local ch_width = math.ceil(width/self.ch_cols)
  local chunks = {}
  for i=1,ch_height do
    chunks[i] = {}
  end
  
  local load_timer = timer:new()

  return setmetatable({ tiles = tiles,
                        layer_data = {},
                        width = width,
                        height = height,
                        num_tiles = width * height,
                        tiles = tiles,
                        ch_height = ch_height,
                        ch_width = ch_width,
                        chunks = chunks,
                        num_chunks = ch_height * ch_width,
                        load_data = load_data,
                        load_timer = load_timer,
                        palette = {}}, tile_map_mt)
end

------------------------------------------------------------------------------
function tile_map:update(dt)
  if self.is_loading then 
  	self:_load_update(dt)
  	self.time_loading = self.load_timer:time_elapsed() / 1000
  end
  if not self.is_loaded then return end
  
  -- calculate which chunks can be seen by the camera
  local topleft = self.camera:get_pos()
  local botright = topleft + vector2:new(self.camera:get_size())
  local mini, minj = self:get_chunk_index(topleft)
  local maxi, maxj = self:get_chunk_index(botright)
  
  mini = math.max(mini, 1)
  minj = math.max(minj, 1)
  maxi = math.min(maxi, self.ch_width)
  maxj = math.min(maxj, self.ch_height)
  
  local visible = true
  if mini > self.ch_width or maxi < 1 then 
    visible = false
  end
  if minj > self.ch_height or maxj < 1 then 
    visible = false
  end
  
  if visible then
    self.map_isvisible = true
    self.min_x, self.min_y = mini, minj
    self.max_x, self.max_y = maxi, maxj
  else
    self.map_isvisible = false
  end
  
  -- update chunks which can be seen by the camera
  if self.map_isvisible then
    for j=self.min_y,self.max_y do
      for i=self.min_x,self.max_x do
        self.chunks[j][i]:update(dt)
      end
    end
  end
  
end


------------------------------------------------------------------------------
function tile_map:draw()
  lg.setColor(C_WHITE)

  local drawcount = 0
  if self.map_isvisible then
    for j=self.min_y,self.max_y do
      for i=self.min_x,self.max_x do
        self.chunks[j][i]:draw()
        drawcount = drawcount + 1
      end
    end
  end
  
end

-----------------------------------------------------------------------------
function tile_map:draw_load_progress()
	local is_loading, stage, percent, time = self:get_load_status()
	
	lg.setColor(C_BLACK)
	msg = "stage: "..stage.." "..percent
	lg.print(msg, 100, 100)
	lg.print("memory: "..math.floor(collectgarbage("count")/1024).." MB", 100,130)
	lg.print("rows, cols = "..rows.." "..cols, 100, 150)
	lg.print("chunks     = "..map.ch_height.." "..map.ch_width, 100, 170)
	lg.print("time       = "..time.."s", 100, 190)
	
	local width = SCR_WIDTH - 100
	local height = 3
	local len
	if percent then
		len = width * percent
	else
		len = width
	end

	lg.setColor(0, 0, 0, 255)
	lg.rectangle("fill", 50, 250, len, height)
	
	local color = self.palette[1][1]
	lg.setColor(color[1], color[2], color[3], 255)
	lg.rectangle("fill", 50, 300, width, 200)
end

------------------------------------------------------------------------------
-- Starts the loading process.
function tile_map:load()
  if self.is_loading then
    return
  end
  self.is_loading = true
  
  -- initialize load_data
  self.load_data.stage = 0
  self.load_data.percent = 0
  self.load_data.context = {1, 1}
  
  self.load_timer:start()
end

------------------------------------------------------------------------------
--[[
   file_name: name of image file
   shades: in the form of {{r, g, b, a}, {r, g, b, a}, {r, g, b, a}}
   tile_type: 'enums' in globals
   
   Layers are processed by priority
     higher index, higher priority
]]--
function tile_map:add_layer(file_name, tile_type, shades)
  if self.is_loaded then
    print("tile_map ERROR: Cannot add layer after map is loaded")
    return
  end  

  self.palette[#self.palette+1] = shades
  self.layer_data[#self.layer_data+1] = {file_name, tile_type, shades}
end

------------------------------------------------------------------------------
-- returns false if not loading
-- returns true, stage_num, percent, total_loading_time if is loading
function tile_map:get_load_status()
  if not self.is_loading then
    return false
  end
  
  return true, self.load_data.stage, self.load_data.percent, self.time_loading
end

------------------------------------------------------------------------------
function tile_map:set_camera(camera)
  self.camera = camera
end

------------------------------------------------------------------------------
-- returns the index of the chunk that contains point
-- returns the chunk if it exists
function tile_map:get_chunk_index(point)
  local width = TILE_WIDTH * self.ch_cols
  local height = TILE_HEIGHT * self.ch_rows
  local i, j = math.floor(point.x / width) + 1, math.floor(point.y / height) + 1
  
  if i >= 1 and i <= self.ch_width and j >= 1 and j <= self.ch_height then
    return i, j, self.chunks[j][i]
  else
    return i, j, false
  end
end


------------------------------------------------------------------------------
-- updates every frame
function tile_map:_load_update(dt)
  local stage = self.load_data.stage

  if     stage == 0 then  -- blank tiles
    local func = function() return self:_load_blank_tiles(5000) end
    if self:_timed_load(func, self.load_time) == false then
      self.load_data.stage = 1
      self.load_data.percent = 0
      self.load_data.context = {1, 1}
    end
    
  elseif stage == 1 then  -- tile neighbours
    local func = function() return self:_load_neighbours(2000) end
    if self:_timed_load(func, self.load_time) == false then
      self.load_data.stage = 2
      self.load_data.percent = 0
      self.load_data.context = {1, 1}
    end
  
  elseif stage == 2 then  -- generate tile images
    if not self.p_imagedata then
      self:_init_imagedata()
    end
    
    -- NOTE: This loader does not seem to work sometimes when not loaded all in
    --       one frame. Until I understand why, we'll just load all images at once.
    --       May not be an issue, this loader stage is pretty quick.
    local func = function() return self:_load_imagedata(self.p_width*self.p_height) end
    if self:_timed_load(func, self.load_time) == false then
      -- create quads for the generated image
      self.p_img = lg.newImage(self.p_imgdata)
      self:_init_palette_quads()
      
      self.load_data.stage = 3
      self.load_data.percent = 0
      self.load_data.context = {1, 1}
    end 
    
  elseif stage == 3 then  -- set tile types
    if not self.layer_imgs then
      self:_init_layer_imgdata()
    end
  
    local func = function() return self:_load_tile_types(2000) end
    if self:_timed_load(func, self.load_time) == false then
      self.load_data.stage = 4
      self.load_data.percent = 0
      self.load_data.context = {1, 1}
    end 
    
  elseif stage == 4 then  -- chunks
  local func = function() return self:_load_chunks(30) end
  if self:_timed_load(func, self.load_time) == false then
    self.load_data.stage = 5
    self.load_data.percent = 0
    self.load_data.context = {1, 1}
  end  
    
  else
    self.is_loading = false
    self.is_loaded = true
  end
end


-----------------------------------------------------------------------------
-- runs a loader function until time t runs out
-- returns true if there is still more to load on the func, false otherwise
function tile_map:_timed_load(func, t)
  local time_left = t
  local start_time = love.timer.getMicroTime()
  
  while time_left > 0 do
    local result = func()
    if not func() then
      return false
    end
    time_left = t - (love.timer.getMicroTime() - start_time)
  end
  
  return true
end

------------------------------------------------------------------------------
--[[ generates the next num number of blank tiles for the map
   returns true if there are still more tiles to load, false otherwise
   
   stage number: 0
   context format: {next_tilecol, next_tilerow}
]]--
function tile_map:_load_blank_tiles(num)
  -- load context
  local next_x, next_y = self.load_data.context[1], self.load_data.context[2]
  
  -- fill in num number of blank tiles
  local count = 0
  local has_next = true
  while count < num and has_next do
    local tile = tile:new()
    self.tiles[next_y][next_x] = tile
    tile.x = TILE_WIDTH * (next_x-1)
    tile.y = TILE_HEIGHT * (next_y-1)
    
    has_next, next_x, next_y = self:_get_next_index(next_x, next_y, self.width, self.height)
    count = count + 1
  end
  
  if has_next then
    -- save context
    self.load_data.percent = (next_y * self.width) / self.num_tiles
    self.load_data.context = {next_x, next_y}
    return true
  else
    return false
  end
  
end

------------------------------------------------------------------------------
--[[ sets neighbours for next num number of tiles int he map
   returns true if there are still more tiles to set, false otherwise
   
   stage number: 1
   context format: {next_tilecol, next_tilerow}
]]--
function tile_map:_load_neighbours(num)
    -- load context
  local next_x, next_y = self.load_data.context[1], self.load_data.context[2]
  
  -- fill in num number of blank tiles
  local count = 0
  local has_next = true
  while count < num and has_next do
    self:_set_neighbours(next_x, next_y)
    
    has_next, next_x, next_y = self:_get_next_index(next_x, next_y, self.width, self.height)
    count = count + 1
  end
  
  if has_next then
    -- save context
    self.load_data.percent = (next_y * self.width) / self.num_tiles
    self.load_data.context = {next_x, next_y}
    return true
  else
    return false
  end
  
end


------------------------------------------------------------------------------
--[[ sets chunks for next num number of chunks for the map
   returns true if there are still more chunks to set, false otherwise
   
   stage number: 2
   context format: {next_chunkcol, next_chunkrow}
]]--
function tile_map:_load_chunks(num)
    -- load context
  local next_x, next_y = self.load_data.context[1], self.load_data.context[2]
  
  -- fill in num number of blank tiles
  local count = 0
  local has_next = true
  while count < num and has_next do
    local chunk = self:_generate_chunk(next_x, next_y)
    self.chunks[next_y][next_x] = chunk
    
    has_next, 
    next_x, next_y = self:_get_next_index(next_x, next_y, self.ch_width, self.ch_height)
    count = count + 1
  end
  
  if has_next then
    -- save context
    self.load_data.percent = (next_y * self.ch_width) / self.num_chunks
    self.load_data.context = {next_x, next_y}
    return true
  else
    return false
  end
end

------------------------------------------------------------------------------
--[[ 
   generates imagedata for the next num tiles
   returns true if there are still images to generate, false otherwise
   
   stage number: 3
   context format: {next_column, next_row}
]]--
function tile_map:_load_imagedata(num)
  -- load context
  local next_x, next_y = self.load_data.context[1], self.load_data.context[2]
  
  -- fill in num number of blank tiles
  local count = 0
  local has_next = true
  local image_data = self.p_imgdata
  while count < num and has_next do
  
    -- if the color exists, generate tile and paste into palette imagedata
    local tile_color = self.palette[next_y][next_x]
    if tile_color then
      local tile_imgdata = self:_generate_tile_imgdata(tile_color, true, C_BORDER)
      
      image_data:paste(tile_imgdata, TILE_WIDTH * (next_x - 1),
                                     TILE_HEIGHT * (next_y - 1), 0, 0,
                                     TILE_WIDTH, TILE_HEIGHT)
    end
    
    has_next, 
    next_x, next_y = self:_get_next_index(next_x, next_y, self.p_width, self.p_height)
    count = count + 1
  end
  
  if has_next then
    -- save context
    local num_shades = self.p_width * self.p_height
    self.load_data.percent = (next_y * self.p_width) / num_shades
    self.load_data.context = {next_x, next_y}
    return true
  else
    return false
  end
end


------------------------------------------------------------------------------
--[[ 
   sets the next num number of tile types
   returns true if there are still more tiles to set, false otherwise
   
   stage number: 4
   context format: {next_tilecol, next_tilerow}
]]--
function tile_map:_load_tile_types(num)
  -- load context
  local next_x, next_y = self.load_data.context[1], self.load_data.context[2]
  
  -- fill in num number of blank tiles
  local count = 0
  local has_next = true
  while count < num and has_next do
    self:_set_tile_type(next_x, next_y)
    
    has_next, next_x, next_y = self:_get_next_index(next_x, next_y, self.width, self.height)
    count = count + 1
  end
  
  if has_next then
    -- save context
    self.load_data.percent = (next_y * self.width) / self.num_tiles
    self.load_data.context = {next_x, next_y}
    return true
  else
    return false
  end
end


------------------------------------------------------------------------------
-- sets the neighbours of the tile located at self.tiles[j][i]
function tile_map:_set_neighbours(i, j)
  local tile = self.tiles[j][i]
  local tiles = self.tiles
  local has_right = i+1 <= self.width
  local has_left = i-1 >= 1
  local has_up = j-1 >= 1
  local has_down = j+1 <= self.height
  
  local neighbours = {}
  if has_right and has_left and has_up and has_down then
    neighbours[UP]        = tiles[j-1][i]
    neighbours[UPRIGHT]   = tiles[j-1][i+1]
    neighbours[RIGHT]     = tiles[j][i+1]
    neighbours[DOWNRIGHT] = tiles[j+1][i+1]
    neighbours[DOWN]      = tiles[j+1][i]
    neighbours[DOWNLEFT]  = tiles[j+1][i-1]
    neighbours[LEFT]      = tiles[j][i-1]
    neighbours[UPLEFT]    = tiles[j-1][i-1]
  else
    if has_up                 then neighbours[UP]        = tiles[j-1][i] end
    if has_right and has_up   then neighbours[UPRIGHT]   = tiles[j-1][i+1] end
    if has_right              then neighbours[RIGHT]     = tiles[j][i+1] end
    if has_right and has_down then neighbours[DOWNRIGHT] = tiles[j+1][i+1] end
    if has_down               then neighbours[DOWN]      = tiles[j+1][i] end
    if has_left and has_down  then neighbours[DOWNLEFT]  = tiles[j+1][i-1] end
    if has_left               then neighbours[LEFT]      = tiles[j][i-1] end
    if has_left and has_up    then neighbours[UPLEFT]    = tiles[j-1][i-1] end
  end
  
  tile:set_neighbours(neighbours)
  
end

------------------------------------------------------------------------------
-- generates chunk for row j, column i
function tile_map:_generate_chunk(i, j)
  local chunk = tile_chunk:new()
  chunk:set_parent(self)
  chunk:set_image(self.p_img, self.p_quads)
  chunk.x = (i-1) * self.ch_cols * TILE_WIDTH
  chunk.y = (j-1) * self.ch_rows * TILE_HEIGHT
  
  -- calculate boundaries for which tiles fin into this chunk
  local minx, miny = (i-1) * self.ch_cols + 1, (j-1) * self.ch_rows + 1
  local maxx, maxy = minx + self.ch_cols - 1, miny + self.ch_rows - 1
  if maxx > self.width then maxx = self.width end
  if maxy > self.height then maxy = self.height end
  
  -- put those tiles in the chunk
  local tiles = {}
  local idx, idy = 1, 1
  for y=miny,maxy do
    local row = {}
    idx = 1
    for x=minx,maxx do
      local tile = self.tiles[y][x]
      tile:set_chunk(chunk)
      tile:set_parent(self)
      tile.batch_x = (x - minx) * TILE_WIDTH
      tile.batch_y = (y - miny) * TILE_HEIGHT
      
      row[idx] = tile
      idx = idx + 1
    end
    
    tiles[idy] = row
    idy = idy + 1
  end
  chunk:set_tiles(tiles)
  chunk:init_sprite_batch()
  return chunk
end

-----------------------------------------------------------------------------
-- color = {r, g, b, a}
-- border = true if has border
-- bcolor (border color)
function tile_map:_generate_tile_imgdata(color, border, bcolor)
  local r, g, b, a = color[1], color[2], color[3], color[4]
  local bdr, bdg, bdb, bda
  if border then
    local f = bcolor[4]/255
    local a = 1-f
    bdr = math.floor(f*bcolor[1] + a*r)
    bdg = math.floor(f*bcolor[2] + a*g)
    bdb = math.floor(f*bcolor[3] + a*b)
  else
    bdr, bdg, bdb, bda = r, g, b, a
  end
  
  local imgdata = love.image.newImageData(TILE_WIDTH, TILE_HEIGHT)
  
  -- fill in border
  for i=0,TILE_WIDTH-1 do     -- horizontal
    imgdata:setPixel(i, 0, bdr, bdg, bdb, 255)
    imgdata:setPixel(i, TILE_HEIGHT-1, bdr, bdg, bdb, 255)
  end
  for j=0,TILE_HEIGHT-1 do     -- vertical
    imgdata:setPixel(0, j, bdr, bdg, bdb, 255)
    imgdata:setPixel(TILE_WIDTH-1, j, bdr, bdg, bdb, 255)
  end
  
  -- fill in body
  for j=1,TILE_HEIGHT-2 do
    for i=1,TILE_WIDTH-2 do
      imgdata:setPixel(i, j, r, g, b, a)
    end
  end
  
  return imgdata
end

------------------------------------------------------------------------------
-- sets the tile type of the tile located at self.tiles[j][i] using the RED
-- channel of the image
function tile_map:_set_tile_type(i, j)
  local layer = nil
  local layers = self.layer_imgs
  
  -- get highest priority layer
  local r = nil
  for k=#layers,1,-1 do
    r = layers[k]:getPixel(i-1, j-1)
    if r ~= 255 then
      layer = k
      break
    end
  end
  if not layer then layer = 1 end
  
  -- calculate color index and set tile type
  local idx = self:_color_to_index(r, #self.palette[layer])
  local t_num = self.layer_data[layer][2]
  local t_type = nil
  local quad = self.p_quads[layer][idx]
  local tile = self.tiles[j][i]
  
  if     t_num == T_WALK then
    t_type = t_walk:new()
  elseif t_num == T_WALL then
    t_type = t_wall:new()
  else
    print("tile_map ERROR: tile type not found")
  end
  
  t_type:set_quad(quad)
  tile:set_type(t_type)
  
end

------------------------------------------------------------------------------
function tile_map:_init_imagedata()
  -- find height
  local height = #self.palette
  if height == 0 then 
    return 
  end
  
  -- find width (maximum shades of a color in the palette)
  local width = 0
  local count = 0
  for i=1,#self.palette do
    shades = self.palette[i]
    width = math.max(width, #shades)
    count = count + #shades
  end
  
  local imgdata = love.image.newImageData(width * TILE_WIDTH, height * TILE_HEIGHT)
  
  self.p_imgdata = imgdata
  self.p_height = height
  self.p_width = width
  self.num_shades = count
  self.tile_images = tile_images
end

------------------------------------------------------------------------------
-- splits the p_quads image into TILE_WIDTH by TILE_HEIGHT pixel quads
function tile_map:_init_palette_quads()
  local image = self.p_img
  local palette = self.palette
  local w, h = TILE_WIDTH, TILE_HEIGHT
  local iw, ih = image:getWidth(), image:getHeight()
  
  local quads = {}
  for j=1,#palette do
    local row = {}
    for i=1,#palette[j] do
      row[i] = lg.newQuad((i-1)*w, (j-1)*h, w, h, iw, ih)
    end
    quads[j] = row
  end
  
  self.p_quads = quads
end


------------------------------------------------------------------------------
-- sets up imgdata from self.layer_data
function tile_map:_init_layer_imgdata()
  local width, height = self.width, self.height
  local layers = self.layer_data
  local layer_imgs = {}
  for i=1,#layers do
    local file = layers[i][1]
    local img = love.image.newImageData(file)
    if img:getWidth() ~= width or img:getWidth() ~= height then
      print("tile_map ERROR: image size does not match map size for file "..file)
    end
    
    layer_imgs[i] = img
  end
  
  self.layer_imgs = layer_imgs
  self.num_layers = #layer_imgs
end


------------------------------------------------------------------------------
-- returns the true, next row, col index of a 2d array
-- returns false if there is not a next index
function tile_map:_get_next_index(col, row, width, height)
  local next_x = col + 1
  local next_y = nil
 
  if next_x > width then
    next_x, next_y = 1, row + 1
    if next_y > height then
      return false
    end
  else
    next_y = row
  end
  
  return true, next_x, next_y
end

------------------------------------------------------------------------------
-- converts a color (0-255) to a shade index (1-num) 
function tile_map:_color_to_index(color, num)
  return math.floor((255 - color)*(num - 1) / 255) + 1
end


return tile_map











